''' Python Code for Communication with the Ecobee Thermostat '''
import requests
import json
import os
import logging

from requests.exceptions import RequestException;

from .const import (ECOBEE_ACCESS_TOKEN,
                    ECOBEE_API_KEY,
                    ECOBEE_AUTHORIZATION_CODE,
                    ECOBEE_CONFIG_FILENAME,
                    ECOBEE_REFRESH_TOKEN)
from .errors import ExpiredTokenError

logger = logging.getLogger('pyecobee')


def config_from_file(filename, config=None):
    ''' Small configuration file management function'''
    if config:
        # We're writing configuration
        try:
            with open(filename, 'w') as fdesc:
                fdesc.write(json.dumps(config))
        except IOError as error:
            logger.exception(error)
            return False
        return True
    else:
        # We're reading config
        if os.path.isfile(filename):
            try:
                with open(filename, 'r') as fdesc:
                    return json.loads(fdesc.read())
            except IOError as error:
                return False
        else:
            return {}


class Ecobee(object):
    ''' Class for storing Ecobee Thermostats and Sensors '''

    def __init__(self, config_filename=None, api_key=None, config=None):
        self.thermostats = None
        self.pin = None
        self.authenticated = False
        self.api_key = None
        self.access_token = None
        self.authorization_code = None
        self.refresh_token = None

        if config is None:
            self.file_based_config = True
            if config_filename is None:
                if api_key is None:
                    logger.error("Error. No API Key was supplied.")
                    return
                jsonconfig = {ECOBEE_API_KEY: api_key}
                config_filename = ECOBEE_CONFIG_FILENAME
                config_from_file(config_filename, jsonconfig)
            config = config_from_file(config_filename)
        else:
            self.file_based_config = False
        self.api_key = config[ECOBEE_API_KEY]
        self.config_filename = config_filename

        if ECOBEE_ACCESS_TOKEN in config:
            self.access_token = config[ECOBEE_ACCESS_TOKEN]

        if ECOBEE_AUTHORIZATION_CODE in config:
            self.authorization_code = config[ECOBEE_AUTHORIZATION_CODE]

        if ECOBEE_REFRESH_TOKEN in config:
            self.refresh_token = config[ECOBEE_REFRESH_TOKEN]

    def request_pin(self) -> bool:
        ''' Method to request a PIN from ecobee for authorization '''
        url = 'https://api.ecobee.com/authorize'
        params = {'response_type': 'ecobeePin',
                  'client_id': self.api_key, 'scope': 'smartWrite'}
        try:
            request = requests.get(url, params=params)
        except RequestException:
            logger.warn("Error connecting to Ecobee.  Possible connectivity outage."
                        "Could not request pin.")
            return False
        if request.status_code == requests.codes.ok:
            self.authorization_code = request.json()['code']
            self.pin = request.json()['ecobeePin']
            logger.info('Please authorize your ecobee developer app with PIN code '
                  + self.pin + '\nGoto https://www.ecobee.com/consumerportal'
                  '/index.html, click\nMy Apps, Add application, Enter Pin'
                  ' and click Authorize.\nAfter authorizing, call request_'
                  'tokens() method.')
            return True
        else:
            logger.warn('Error while requesting pin from ecobee.com.'
                        ' Status code: ' + str(request.status_code))
            return False

    def request_tokens(self) -> bool:
        ''' Method to request API tokens from ecobee '''
        url = 'https://api.ecobee.com/token'
        params = {'grant_type': 'ecobeePin', 'code': self.authorization_code,
                  'client_id': self.api_key}
        try:
            request = requests.post(url, params=params)
        except RequestException:
            logger.warn("Error connecting to Ecobee.  Possible connectivity outage."
                        "Could not request token.")
            return False
        if request.status_code == requests.codes.ok:
            self.access_token = request.json()['access_token']
            self.refresh_token = request.json()['refresh_token']
            self.write_tokens_to_file()
            self.pin = None
            return True
        else:
            logger.warn('Error while requesting tokens from ecobee.com.'
                  ' Status code: ' + str(request.status_code))
            return False

    def refresh_tokens(self) -> bool:
        ''' Method to refresh API tokens from ecobee '''
        url = 'https://api.ecobee.com/token'
        params = {'grant_type': 'refresh_token',
                  'refresh_token': self.refresh_token,
                  'client_id': self.api_key}
        try:
            request = requests.post(url, params=params)
        except RequestException:
            logger.warn("Error connecting to Ecobee.  Possible connectivity outage."
                        "Could not refresh token.")
            return False
        if request.status_code == requests.codes.ok:
            self.access_token = request.json()['access_token']
            self.refresh_token = request.json()['refresh_token']
            self.write_tokens_to_file()
            return True
        else:
            logger.warn('Error while refreshing tokens from ecobee.com.'
                  ' Status code: ' + str(request.status_code))
            return False

    def get_thermostats(self):
        ''' Set self.thermostats to a json list of thermostats from ecobee '''
        url = 'https://api.ecobee.com/1/thermostat'
        header = {'Content-Type': 'application/json;charset=UTF-8',
                  'Authorization': 'Bearer ' + self.access_token}
        params = {'json': ('{"selection":{"selectionType":"registered",'
                            '"includeRuntime":"true",'
                            '"includeSensors":"true",'
                            '"includeProgram":"true",'
                            '"includeEquipmentStatus":"true",'
                            '"includeEvents":"true",'
                            '"includeWeather":"true",'
                            '"includeSettings":"true"}}')}
        try:
            request = requests.get(url, headers=header, params=params)
        except RequestException:
            logger.warn("Error connecting to Ecobee.  Possible connectivity outage.")
            return None
        if request.status_code == requests.codes.ok:
            self.authenticated = True
            self.thermostats = request.json()['thermostatList']
            return self.thermostats
        elif request.status_code in [400, 401]:
            self.authenticated = False
            raise ExpiredTokenError("Tokens have expired. Request new tokens from ecobee.")
        else:
            logger.info("Error connecting to Ecobee while attempting to get "
                  "thermostat data.  ")
            return None

    def get_thermostat(self, index):
        ''' Return a single thermostat based on index '''
        return self.thermostats[index]

    def get_remote_sensors(self, index):
        ''' Return remote sensors based on index '''
        return self.thermostats[index]['remoteSensors']

    def write_tokens_to_file(self):
        ''' Write api tokens to a file '''
        config = dict()
        config[ECOBEE_API_KEY] = self.api_key
        config[ECOBEE_ACCESS_TOKEN] = self.access_token
        config[ECOBEE_REFRESH_TOKEN] = self.refresh_token
        config[ECOBEE_AUTHORIZATION_CODE] = self.authorization_code
        if self.file_based_config:
            config_from_file(self.config_filename, config)
        else:
            self.config = config

    def update(self):
        ''' Get new thermostat data from ecobee '''
        self.get_thermostats()

    def make_request(self, body, log_msg_action, *, retry_count=0):
        url = 'https://api.ecobee.com/1/thermostat'
        header = {'Content-Type': 'application/json;charset=UTF-8',
                  'Authorization': 'Bearer ' + self.access_token}
        params = {'format': 'json'}
        try:
            request = requests.post(url, headers=header, params=params, json=body)
        except RequestException:
            logger.warn("Error connecting to Ecobee.  Possible connectivity outage.")
            return None
        if request.status_code == requests.codes.ok:
            return request
        elif request.status_code in [400, 401] and retry_count == 0:
            if self.refresh_tokens():
                return self.make_request(
                    body, log_msg_action, retry_count=retry_count + 1
                )
        else:
            logger.info(
                "Error fetching data from Ecobee while attempting to %s: %s",
                log_msg_action, request.json())
            return None

    def set_hvac_mode(self, index, hvac_mode):
        ''' possible hvac modes are auto, auxHeatOnly, cool, heat, off '''
        body = {"selection": {"selectionType": "thermostats",
                              "selectionMatch": self.thermostats[index]['identifier']},
                              "thermostat": {
                                  "settings": {
                                      "hvacMode": hvac_mode
                                  }
                              }}
        log_msg_action = "set HVAC mode"
        return self.make_request(body, log_msg_action)

    def set_fan_min_on_time(self, index, fan_min_on_time):
        ''' The minimum time, in minutes, to run the fan each hour. Value from 1 to 60 '''
        body = {"selection": {"selectionType": "thermostats",
                        "selectionMatch": self.thermostats[index]['identifier']},
                        "thermostat": {
                            "settings": {
                                "fanMinOnTime": fan_min_on_time
                            }
                        }}
        log_msg_action = "set fan minimum on time."
        return self.make_request(body, log_msg_action)

    def set_fan_mode(self, index, fan_mode, cool_temp, heat_temp, hold_type="nextTransition"):
        ''' Set fan mode. Values: auto, minontime, on '''
        body = {"selection": {
                    "selectionType": "thermostats",
                    "selectionMatch": self.thermostats[index]['identifier']},
                "functions": [{"type": "setHold", "params": {
                    "holdType": hold_type,
                    "coolHoldTemp": int(cool_temp * 10),
                    "heatHoldTemp": int(heat_temp * 10),
                    "fan": fan_mode
                }}]}
        log_msg_action = "set fan mode"
        return self.make_request(body, log_msg_action)

    def set_hold_temp(self, index, cool_temp, heat_temp,
                      hold_type="nextTransition"):
        ''' Set a hold '''
        body = {"selection": {
                    "selectionType": "thermostats",
                    "selectionMatch": self.thermostats[index]['identifier']},
                "functions": [{"type": "setHold", "params": {
                    "holdType": hold_type,
                    "coolHoldTemp": int(cool_temp * 10),
                    "heatHoldTemp": int(heat_temp * 10)
                }}]}
        log_msg_action = "set hold temp"
        return self.make_request(body, log_msg_action)

    def set_climate_hold(self, index, climate, hold_type="nextTransition"):
        ''' Set a climate hold - ie away, home, sleep '''
        body = {"selection": {
                    "selectionType": "thermostats",
                    "selectionMatch": self.thermostats[index]['identifier']},
                "functions": [{"type": "setHold", "params": {
                    "holdType": hold_type,
                    "holdClimateRef": climate
                }}]}
        log_msg_action = "set climate hold"
        return self.make_request(body, log_msg_action)
      
    def create_vacation(self, index, vacation_name, cool_temp, heat_temp, start_date=None, start_time=None,
                     end_date=None, end_time=None, fan_mode="auto", fan_min_on_time="0"):
        ''' Create a vacation hold'''
        body = {"selection": {
                    "selectionType": "thermostats",
                    "selectionMatch": self.thermostats[index]['identifier']},
                "functions": [{"type": "createVacation", "params": {
                    "name": vacation_name,
                    "coolHoldTemp": int(cool_temp * 10),
                    "heatHoldTemp": int(heat_temp * 10),
                    "startDate": start_date,
                    "startTime": start_time,
                    "endDate": end_date,
                    "endTime": end_time,
                    "fan_mode": fan_mode,
                    "fan_min_on_time": fan_min_on_time
                }}]}
        log_msg_action = "create vacation hold"
        return self.make_request(body, log_msg_action)

    def delete_vacation(self, index, vacation):
        ''' Delete the vacation with name vacation '''
        body = {"selection": {
                    "selectionType": "thermostats",
                    "selectionMatch": self.thermostats[index]['identifier']},
                "functions": [{"type": "deleteVacation", "params": {
                    "name": vacation
                }}]}

        log_msg_action = "delete a vacation"
        return self.make_request(body, log_msg_action)

    def resume_program(self, index, resume_all=False):
        ''' Resume currently scheduled program '''
        body = {"selection": {
                    "selectionType": "thermostats",
                    "selectionMatch": self.thermostats[index]['identifier']},
                "functions": [{"type": "resumeProgram", "params": {
                    "resumeAll": resume_all
                }}]}

        log_msg_action = "resume program"
        return self.make_request(body, log_msg_action)

    def send_message(self, index, message="Hello from python-ecobee!"):
        ''' Send a message to the thermostat '''
        body = {"selection": {
                    "selectionType": "thermostats",
                    "selectionMatch": self.thermostats[index]['identifier']},
                "functions": [{"type": "sendMessage", "params": {
                    "text": message[0:500]
                }}]}

        log_msg_action = "send message"
        return self.make_request(body, log_msg_action)

    def set_humidity(self, index, humidity):
        ''' Set humidity level'''
        body = {"selection": {"selectionType": "thermostats",
                              "selectionMatch": self.thermostats[index]['identifier']},
                              "thermostat": {
                                  "settings": {
                                      "humidity": int(humidity)
                                  }
                              }}

        log_msg_action = "set humidity level"
        return self.make_request(body, log_msg_action)

    def set_mic_mode(self, index, mic_enabled):
        '''Enable/disable Alexa mic (only for Ecobee 4)
        Values: True, False
        '''

        body = {
            'selection': {
                'selectionType': 'thermostats',
                'selectionMatch': self.thermostats[index]['identifier']},
            'thermostat': {
                'audio': {
                    'microphoneEnabled': mic_enabled}}}

        log_msg_action = 'set mic mode'
        return self.make_request(body, log_msg_action)

    def set_occupancy_modes(self, index, auto_away=None, follow_me=None):
        '''Enable/disable Smart Home/Away and Follow Me modes
        Values: True, False
        '''

        body = {
            'selection': {
                'selectionType': 'thermostats',
                'selectionMatch': self.thermostats[index]['identifier']},
            'thermostat': {
                'settings': {
                    'autoAway': auto_away,
                    'followMeComfort': follow_me}}}

        log_msg_action = 'set occupancy modes'
        return self.make_request(body, log_msg_action)

    def set_dst_mode(self, index, dst):
        '''Enable/disable daylight savings
        Values: True, False
        '''

        body = {
            'selection': {
                'selectionType': 'thermostats',
                'selectionMatch': self.thermostats[index]['identifier']},
            'thermostat': {
                'location': {
                    'isDaylightSaving': dst}}}

        log_msg_action = 'set dst mode'
        return self.make_request(body, log_msg_action)
